#include "dfu_bluetooth.h"

#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/sys/reboot.h>
#include <zephyr/mgmt/mcumgr/mgmt/mgmt.h>
#include <zephyr/mgmt/mcumgr/mgmt/callbacks.h>
#include <zephyr/dfu/mcuboot.h>
#include <zephyr/logging/log.h>
#include <zephyr/settings/settings.h>

#include "peri_led.h"

LOG_MODULE_REGISTER(dfu_bluetooth, 4);

static const struct bt_data ad[] = {
    BT_DATA_BYTES(BT_DATA_FLAGS, (BT_LE_AD_GENERAL | BT_LE_AD_NO_BREDR)),
    BT_DATA(BT_DATA_NAME_COMPLETE, "ECG_DFU", 7),
};

static const struct bt_data sd[] = {
    BT_DATA_BYTES(BT_DATA_UUID128_ALL,
                  0x84, 0xaa, 0x60, 0x74, 0x52, 0x8a, 0x8b, 0x86,
                  0xd3, 0x4c, 0xb7, 0x1d, 0x1d, 0xdc, 0x53, 0x8d),
};

static void led_timer_cb(struct k_timer *timer_id);
K_TIMER_DEFINE(led_timer, led_timer_cb, NULL);
static void led_timer_cb(struct k_timer *timer_id)
{
    ARG_UNUSED(timer_id);
    led_toggle();
    k_timer_start(&led_timer, K_MSEC(500), Z_TIMEOUT_NO_WAIT);
}

void dfu_enter()
{
    int dfu_mode = 1;

    settings_save_one("dfu", &dfu_mode, sizeof(dfu_mode));
    settings_save();

    LOG_INF("Reboot to enter DFU mode");
    
    k_sleep(K_MSEC(500));
    sys_reboot(SYS_REBOOT_COLD);
}

struct k_work_delayable dfu_not_started_handler;

static enum mgmt_cb_return dfu_mode_cb(uint32_t event, enum mgmt_cb_return prev_status,
				       int32_t *rc, uint16_t *group, bool *abort_more, void *data,
				       size_t data_size)
{
    switch (event)
    {
    case MGMT_EVT_OP_IMG_MGMT_DFU_STARTED:
        LOG_INF("DFU Started");
        k_work_reschedule(&dfu_not_started_handler, K_MINUTES(5));
        break;

    case MGMT_EVT_OP_IMG_MGMT_DFU_STOPPED:
        LOG_INF("DFU Stopped");
        sys_reboot(SYS_REBOOT_COLD);
        break;

    case MGMT_EVT_OP_IMG_MGMT_DFU_PENDING:
        LOG_INF("DFU Pending");
        break;
        
    case MGMT_EVT_OP_IMG_MGMT_DFU_CONFIRMED:
        LOG_INF("DFU Confirmed");
        break;

    default:
        LOG_ERR("Unknown event %d", event);
        break;
    }

    return MGMT_ERR_EOK;
}

static struct mgmt_callback dfu_mode_mgmt_cb = {
    .callback = dfu_mode_cb,
    .event_id = MGMT_EVT_OP_IMG_MGMT_DFU_STARTED | MGMT_EVT_OP_IMG_MGMT_DFU_STOPPED |
                MGMT_EVT_OP_IMG_MGMT_DFU_PENDING | MGMT_EVT_OP_IMG_MGMT_DFU_CONFIRMED,
};

static void exit_dfu_mode(struct k_work *work)
{
    LOG_WRN("DFU did not start or could not complete. Reset to exit dfu mode");
    k_sleep(K_MSEC(1000));
    sys_reboot(SYS_REBOOT_COLD);
}

void init_dfu_bluetooth()
{
    // LED initialization moved to @ref peri_led

    int err = bt_enable(NULL);
    if (err)
    {
        LOG_ERR("Bluetooth init failed (err %d)\n", err);
        return;
    }

    mgmt_callback_register(&dfu_mode_mgmt_cb);
    
    settings_load();

    err = bt_le_adv_start(BT_LE_ADV_CONN, ad, ARRAY_SIZE(ad),
                          sd, ARRAY_SIZE(sd));
    LOG_INF("DFU adv started with err: %d", err);

    k_work_init_delayable(&dfu_not_started_handler, exit_dfu_mode);
    k_work_reschedule(&dfu_not_started_handler, K_MINUTES(5));

    k_timer_start(&led_timer, K_MSEC(500), Z_TIMEOUT_NO_WAIT);

    LOG_INF("Bluetooth DFU initialized\n");
}

void dfu_check_boot_image()
{
    if (boot_is_img_confirmed())
    {
        return;
    }

    int ret = boot_write_img_confirmed();
    if (ret)
    {
        LOG_ERR("Couldn't confirm image: %d", ret);
    }
    else
    {
        LOG_INF("Marked image as OK");
    }
}